/*******************************************************\
* modlisation d'une surface d'eau                      *
* WaterSurface3.cpp                                     *
* ------------------------------------------------------*
* Georges AZZI / Georges CAMY                           * 
* Julie POULETTY / Arnaud QUENUM                        *
* IMAC Ing3 2002/2003 - Universit de Marne La Valle  *
* Copyright 2003                                        *
\*******************************************************/
#include "WaterSurface3.h"


/*******************************************************\
*                    constructor                       *
\*******************************************************/
WaterSurface3::WaterSurface3(GLint newLength, GLint newWidth){
	//set the grid size (isn't dynamic yet, values must be 128x128)
	length = newLength;
	width = newWidth;
	
	// initialize height field (z)
	for (int i=0; i<length; i++){
		for (int j=0; j<width; j++){
			z[i][j] = 0;	
		}
	}
				
	// set elapsed time (used in update)
	time = glutGet(GLUT_ELAPSED_TIME);

    // define constants from Pierson-Moskowitz model
    alpha = 0.0081;         // constant from Pierson-Moskowitz model 
    omega = 10.0f;          // surface size
    g = 9.80665;            // gravity
    nWaves = 1;             // number of calculated waves
    maxwave = 15;            // maximum number of pre-calculated waves
    currentWave = maxwave;  // current wave. (current wave == maxwave is the combined wave)
    windSpeed = 5.5f;       // windspeed
    windDirection = 0.1f;   // winddirection

    // precalculate the waves
    srand((int)time);
    reset();
}

/*******************************************************\
  Mthode:	Set / Get				 
  Description:  windspeed
                winddirection
                windnumber
\*******************************************************/
void WaterSurface3::SetWindSpeed(float speed){
    windSpeed = speed;
}

float WaterSurface3::GetWindSpeed(){
    return windSpeed;
}

void WaterSurface3::SetWindDirection(float direction){
    windDirection = direction;
}

float WaterSurface3::GetWindDirection(){
    return windDirection;
}

void WaterSurface3::SetCurrentWave(int w){
    if (w>maxwave) w=1;
    currentWave = w;
}

int WaterSurface3::GetCurrentWave(){
    return currentWave;
}



/*******************************************************\
  Mthode:	Gamma				 
  Description:  some sort of generalized faculty 
   		function	
\*******************************************************/
double WaterSurface3::Gamma (double x){
   x = x-1;             
   double faculty = 1.0;            
   double upperlimit = x;           
   
   if (floor(x)==x){ 
       upperlimit = x + 1.0;  
   }
   
   for(int i=1; i<upperlimit; i++){
      faculty = faculty * i;
   }                  
   double fact = floor(x)*x + 1.0 - floor(x)*floor(x); 
   faculty = faculty * fact;
   return faculty;
}



/*******************************************************\
  Mthode:	    MitsuyasuDistribution				 
  Description:  gives the energy of a wave that propagates
  		with frequency f and angle theta with 
  		respect to the wind direction	
\*******************************************************/
double WaterSurface3::MitsuyasuDistribution (double f, double theta) {
   double s;
   double ang = cos(theta/2);
   ang = ang*ang;                   // cos(theta/2)^2
   if (f<fPeak){                    // maximum frequency
      s = 9.77 * pow(f/fPeak, -2.5);
   } 
   else {
      s = 6.97 * pow(f/fPeak, 5.0);
   }
   return 5* ( Gamma(s+1) * pow(ang,s)/(2*sqrt(M_PI) * Gamma(s+0.5)) );
}


/*******************************************************\
  Mthode:	    EPM				 
  Description:  Energy distribution from the 
  		Pierson-Moskowitz model. Gives the 
  		distribution for a given frequency.
\*******************************************************/
double WaterSurface3::EPM(double f){
   double EPM;    
   EPM = alpha * g * g * exp(-1.25 * pow(f/fPeak,-4))/pow(2*M_PI,4);
   return EPM;
}



/*******************************************************\
  Mthode:	    EnergyDistribution				 
  Description:  The energy distribution from both the 
  		angular distribution and de Pierson-Moskowitz
  		model. f is the frequency and theta the angle
  		with the wind.
\*******************************************************/
double WaterSurface3::EnergyDistribution(double f, double theta){
   return EPM(f) * MitsuyasuDistribution(f,theta);
}



/*******************************************************\
  Mthode:	    Amplitude				 
  Description:  The amplitude of the wave, considering 
  		the frequency, the angle with de wind, and 
  		the direction of the wave.
\*******************************************************/
double WaterSurface3::Amplitude(double f, double theta, double k){
   return sqrt(EnergyDistribution(f,theta) * g * M_PI * M_PI/(k*f*omega));
}



/*******************************************************\
  Mthode:	    Reset				 
  Description:  recalculates the waves
\*******************************************************/
void WaterSurface3::reset()
{
   fPeak = 0.13 * g/windSpeed;     //compute the peak frequency
   WAVE *w = &wave[0];
   nWaves=0;                       // number of calculated waves
   int timelimit=0;                // don't search endlessly...

   while (nWaves<maxwave) {
      float f = fPeak + rand() / 16384.0f * 0.5 - 0.5;

      if (f>0.0) {
          w->lambda      = 2.0*M_PI/(pow( 2.0*M_PI*f,2.0)/g);
          w->k           = 2.0*M_PI/w->lambda;
          w->omega       = sqrt(g*w->k);
          w->freq        = w->omega / (2.0*M_PI);
          w->periode     = 1.0f/w->freq;
          w->direction   = (rand()/16384.0f+0.5f)*M_PI;
          w->dirX        = cos(w->direction)*0.5f;
          w->dirY        = sin(w->direction)*0.5f;

          double phi0    = rand()/16384.0f*M_PI;
          double a0      = Amplitude(w->freq,w->direction-windDirection,w->k);

          w->amplitude   = a0*cos( phi0 );
          w->phase       = a0*sin( phi0 );

          
          if (fabs(w->amplitude)>=0.001f ) {
                w ++;
                nWaves ++;
		  }
	  }
   }
}



/*******************************************************\
  Mthode:		update						
  Description:  calculates the next heightvalues	
\*******************************************************/
void WaterSurface3::update(){
	time = glutGet(GLUT_ELAPSED_TIME)/500.0f;

	for ( int j = 0; j < length; j++ )
		for ( int i = 0; i < width; i++ ){
			float h = 0.0f;

            // afficher toutes les vagues
            if (currentWave == maxwave){
                for ( int k = 0; k < nWaves; k++ ){
				    h += wave[ k ].amplitude * 
					    (float)sin( wave[ k ].phase + wave[ k ].omega * wave[ k ].freq * ( time +
					    wave[ k ].dirX * i + 
					    wave[ k ].dirY * j ) );
                    time = time + 0.00002f;
			    }
            }
            // n'afficher que la vague numro currentWave
            else{
				    h = wave[ currentWave ].amplitude * 
					    (float)sin( wave[ currentWave].phase + wave[ currentWave ].omega * wave[ currentWave ].freq * ( time +
					    wave[ currentWave ].dirX * i + 
					    wave[ currentWave ].dirY * j) );
            }
            z[i][j] = h;
		}
}



/*******************************************************\
  Mthode:	    dessiner								
  Description:  draws the waves.                
\*******************************************************/
void WaterSurface3::dessiner(){
	
	glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
	glDisable(GL_LIGHTING);
	glColor3f(0.0,0.65,0.8);
	
	glPushMatrix();

	for (int i=1;i<length-1;i++){
		glBegin(GL_QUAD_STRIP);
		for (int j=1;j<width-1;j++){
			// dessiner les points
			glVertex3d(i,j,z[i][j]);
			glVertex3d(i+1,j,z[i+1][j]);
		}
		glEnd();
	}

	glPopMatrix();

	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}



/*******************************************************\
  Mthode:		PrintData						
  Description:  affiche les donnes de la vague	
\*******************************************************/
void WaterSurface3::PrintData()
{
    /*
    for ( int k = 0; k < nWaves; k++ ){
        cout<< "// Wave " << k <<endl
            << "wave[" << k <<"].amplitude = "<< wave[k].amplitude << ";" <<endl
            << "wave[" << k <<"].phase = "<< wave[k].phase << ";" <<endl
            << "wave[" << k <<"].omega = "<< wave[k].omega << ";" <<endl
            << "wave[" << k <<"].freq = "<< wave[k].freq << ";" <<endl
            << "wave[" << k <<"].dirX = "<< wave[k].dirX << ";" <<endl
            << "wave[" << k <<"].dirY = "<< wave[k].dirY << ";" <<endl;
    }
    */
    for ( int k = 0; k < nWaves; k++ ){
        cout<< "// Wave " << k <<endl
            << "float wave" << k <<"Amplitude = "<< wave[k].amplitude << ";" <<endl
            << "float wave" << k <<"Phase = "<< wave[k].phase << ";" <<endl
            << "float wave" << k <<"Omega = "<< wave[k].omega << ";" <<endl
            << "float wave" << k <<"Freq = "<< wave[k].freq << ";" <<endl
            << "float wave" << k <<"DirX = "<< wave[k].dirX << ";" <<endl
            << "float wave" << k <<"DirY = "<< wave[k].dirY << ";" <<endl;
    }
}